import inspect
from inspect import Parameter
from typing import Type, Set, Dict
from threading import Lock

from ..bindings.provider_registry import ProviderMappings
from ..bindings.registry import DependencyMappings
from ..constants.dependency_scope import DependencyScope
from ..exception.exceptions import (
    UnregisteredDependencyException,
    CircularDependencyException,
    UnspecifiedDependencyTypeException,
)
from ..resolver.singleton_bucket import SingletonBucket
from ..types.generic_types import T
from pinjet.common.constants import Index


class DependencyResolver:

    __lock: Lock = Lock()

    def __new__(cls):
        with DependencyResolver.__lock:
            if not hasattr(DependencyResolver, "__singleton_instance"):
                setattr(cls, '__singleton_instance', super(DependencyResolver, cls).__new__(cls))
        return getattr(cls, '__singleton_instance')

    @staticmethod
    def resolve(clazz: Type[T]) -> T:

        resolver = DependencyResolver.__get_singleton_instance()

        with DependencyResolver.__lock:

            contextual_dependency_set: Set[Type] = set()

            return resolver.__resolve_dependency(clazz, contextual_dependency_set)

    @staticmethod
    def __get_singleton_instance() -> 'DependencyResolver':
        return DependencyResolver()

    def __resolve_dependency(self, clazz: Type[T], contextual_dependency_set: Set[Type]) -> T:

        if self.__is_resolvable(clazz) is not True:
            raise UnregisteredDependencyException(clazz)

        if ProviderMappings.contains_resolver_for(clazz):
            return self.__resolve_by_provided_procedure(clazz, contextual_dependency_set)

        scope: DependencyScope = DependencyMappings.get_scope(clazz)
        implementation = DependencyMappings.get_binding(clazz)

        if implementation in contextual_dependency_set:
            raise CircularDependencyException(implementation)
        else:
            contextual_dependency_set.add(implementation)

        if scope is DependencyScope.SINGLETON and SingletonBucket.contains_singleton(implementation):

            return SingletonBucket.get(implementation)

        dependency_dictionary: Dict[str, Type] = {}

        for parameter_name, parameter_type in inspect.signature(clazz.__init__).parameters.items():

            if self.__parameter_is_skippable(parameter_name, parameter_type):
                continue

            if parameter_type.annotation is Parameter.empty:
                raise UnspecifiedDependencyTypeException(
                    f'Missing type for Constructor parameter: {parameter_name}'
                )

            dependency_dictionary[parameter_name] = self.__resolve_dependency(
                parameter_type.annotation,
                contextual_dependency_set
            )

        resolved = implementation(**dependency_dictionary)

        if scope is DependencyScope.SINGLETON:
            SingletonBucket.put(implementation, resolved)

        contextual_dependency_set.remove(implementation)

        return resolved

    def __resolve_by_provided_procedure(self, clazz: Type[T], contextual_dependency_set: Set[Type]) -> T:

        if clazz in contextual_dependency_set:
            raise CircularDependencyException(clazz)
        else:
            contextual_dependency_set.add(clazz)

        if SingletonBucket.contains_singleton(clazz):
            return SingletonBucket.get(clazz)

        source_class: Type[T] = ProviderMappings.get_source_class(clazz)
        source_class_instance = self.__resolve_dependency(source_class, contextual_dependency_set)

        source_function_name: str = ProviderMappings.get_source_function_name(clazz)
        source_function = [
            (name, member)
            for name, member in inspect.getmembers(source_class_instance)
            if name == source_function_name
        ][Index.ZERO][Index.LAST]

        resolved_instance: T = source_function()

        SingletonBucket.put(clazz, resolved_instance)

        contextual_dependency_set.remove(clazz)

        return resolved_instance

    @staticmethod
    def __is_resolvable(clazz: Type) -> bool:

        if DependencyMappings.contains_binding_for(clazz):
            return True

        if ProviderMappings.contains_resolver_for(clazz):
            return True

        return False

    @staticmethod
    def __parameter_is_skippable(parameter_name: str, parameter_type: Parameter) -> bool:

        if parameter_name == 'self':
            return True

        if parameter_name == 'cls':
            return True

        if parameter_name == 'args' and parameter_type.annotation == Parameter.empty:
            return True

        if parameter_name == 'kwargs' and parameter_type.annotation == Parameter.empty:
            return True

        return False



import inspect
import functools
from typing import List, Callable

from pinjet.common.constants import Index, Integer

from ..exception.exceptions import TargetElementTypeMisMatchException, TargetTypeNotSpecifiedException
from ..constants.element_type import ElementType


def target(declared_target_element_type):

    if declared_target_element_type is None or not isinstance(declared_target_element_type, ElementType):

        raise TargetTypeNotSpecifiedException

    elif (isinstance(declared_target_element_type, ElementType)
          and declared_target_element_type == ElementType.UNDETERMINED):

        raise ValueError('Target type cannot be undetermined')

    def decorator(target_decorator):

        @functools.wraps(target_decorator)
        def wrapper(*args, **kwargs):

            # Check if the target_decorator accepts its own arguments
            if args and callable(args[Index.ZERO]):

                # If args[0] is a callable, it's likely a function or method
                element_type = __detect_element_type(args[Index.ZERO])

                if element_type != declared_target_element_type:
                    raise TargetElementTypeMisMatchException(target_decorator,
                                                             declared_target_element_type,
                                                             element_type)

            else:

                # Check if the target_decorator is applied to an element
                if len(args) > 1:

                    element_type = __detect_element_type(args[Index.ONE])

                    if element_type != declared_target_element_type:
                        raise TargetElementTypeMisMatchException(target_decorator,
                                                                 declared_target_element_type,
                                                                 element_type)

            # Call the target_decorator
            return target_decorator(*args, **kwargs)

        return wrapper

    return decorator


def __is_class_method(element: Callable, args: List[str]) -> bool:

    if isinstance(element, classmethod):
        return True

    if len(args) != Integer.ZERO and args[Index.ZERO] == 'cls':
        return True

    return False


def __is_instance_method(args: List[str]) -> bool:

    if len(args) != Integer.ZERO and args[Index.ZERO] == 'self':
        return True

    return False


def __detect_element_type(element) -> ElementType:

    if isinstance(element, type):

        return ElementType.CLASS

    elif hasattr(element, '__call__'):

        element_arguments = inspect.getfullargspec(element).args

        if __is_class_method(element, element_arguments):

            return ElementType.CLASS_METHOD

        elif __is_instance_method(element_arguments):

            return ElementType.INSTANCE_METHOD

        elif inspect.isfunction(element):

            return ElementType.FUNCTION

        else:

            return ElementType.UNDETERMINED



from typing import Type, Union, Callable

from .target import target
from ..bindings.registry import DependencyMappings
from ..constants.element_type import ElementType
from ..constants.dependency_scope import DependencyScope


@target(ElementType.CLASS)
def injectable(cls: Type = None,
               scope: DependencyScope = DependencyScope.SINGLETON) -> Union[Type, Callable[[Type], Type]]:
    """

    A decorator to register classes as injectable

    :param cls: The class to be registered in the dependency registry
    :param scope: Lifespan of instance - multiple or single
    :return: the registered class or the decorated function that will register the class
    """

    if cls is not None:

        if isinstance(cls, DependencyScope) is False:
            DependencyMappings.bind(cls, cls, scope)
            return cls
        else:
            scope = cls

    def injectable_decorator(clazz: Type) -> Type:

        DependencyMappings.bind(clazz, clazz, scope)

        return clazz

    return injectable_decorator



from threading import Lock
from typing import Dict, Type

from ..exception.exceptions import DuplicateSingletonInstanceException
from ..types.generic_types import T


class SingletonBucket:

    __lock = Lock()

    def __init__(self):
        with SingletonBucket.__lock:
            if not hasattr(self, '_bucket'):
                self._bucket: Dict[Type[T], T] = {}

    def __new__(cls):
        with SingletonBucket.__lock:
            if not hasattr(cls, '__singleton_instance'):
                setattr(cls, '__singleton_instance', super(SingletonBucket, cls).__new__(cls))
        return getattr(cls, '__singleton_instance')

    @staticmethod
    def __get_singleton_instance() -> 'SingletonBucket':
        return SingletonBucket()

    @staticmethod
    def put(key: Type[T], value: T) -> None:
        singleton_bucket = SingletonBucket.__get_singleton_instance()
        with SingletonBucket.__lock:
            if key in singleton_bucket._bucket:
                raise DuplicateSingletonInstanceException(key)
            singleton_bucket._bucket[key] = value

    @staticmethod
    def get(key: Type[T]) -> T:
        return SingletonBucket.__get_singleton_instance()._bucket.get(key)

    @staticmethod
    def contains_singleton(key: Type[T]) -> bool:
        return SingletonBucket.__get_singleton_instance()._bucket.get(key) is not None



from typing import Dict, Type, Union

from ..types.generic_types import T
from ..types.resolution_procedure import ResolutionProcedure


class ProviderMappings:

    def __init__(self):
        if not hasattr(self, '_mappings'):
            self._mappings: Dict[Union[Type, Type[T]], ResolutionProcedure] = {}

    def __new__(cls):
        if not hasattr(cls, '_singleton_instance'):
            setattr(cls, '_singleton_instance', super(ProviderMappings, cls).__new__(cls))
        return getattr(cls, '_singleton_instance')

    @staticmethod
    def __get_singleton_instance() -> 'ProviderMappings':
        return ProviderMappings()

    @staticmethod
    def put(clazz: Union[Type, Type[T]], source: ResolutionProcedure) -> None:
        ProviderMappings.__get_singleton_instance()._mappings[clazz] = source

    @staticmethod
    def get_source_class(clazz: Union[Type, Type[T]]) -> Union[Type, Type[T]]:
        return ProviderMappings.__get_singleton_instance()._mappings.get(clazz).source_class

    @staticmethod
    def get_source_function_name(clazz: Union[Type, Type[T]]) -> str:
        return ProviderMappings.__get_singleton_instance()._mappings.get(clazz).source_function_name

    @staticmethod
    def contains_resolver_for(clazz: Union[Type, Type[T]]) -> bool:
        return True if clazz in ProviderMappings.__get_singleton_instance()._mappings else False



import os
import inspect
from types import ModuleType
from importlib.machinery import ModuleSpec
from importlib.util import spec_from_file_location, module_from_spec
from typing import Callable

from pinjet.common.constants import String
from pinjet.common.constants.io import SKIPPABLE_DIRECTORIES, PYTHON_SCRIPT_EXTENSION


class RepositoryScanner:

    @staticmethod
    def scan_repository(main_function: Callable):

        source_directory: str = os.path.dirname(inspect.getfile(main_function))

        for root, _, files in os.walk(source_directory):

            for file_name in files:

                if not file_name.endswith(PYTHON_SCRIPT_EXTENSION):
                    continue

                file_path = os.path.join(root, file_name)
                module_name = file_name.replace(PYTHON_SCRIPT_EXTENSION, String.EMPTY)

                if any(map(lambda skippable_directory: skippable_directory in file_path, SKIPPABLE_DIRECTORIES)):
                    continue

                loaded_spec: ModuleSpec = spec_from_file_location(module_name, file_path)
                loaded_module: ModuleType = module_from_spec(loaded_spec)
                loaded_spec.loader.exec_module(loaded_module)



from typing import Callable, Type

from pinjet.common.constants import String
from pinjet.core.constants.element_type import ElementType


class UnregisteredDependencyException(Exception):

    def __init__(self, dependency: Type):
        super().__init__(f'Unregistered dependency "{dependency.__qualname__}"')


class CircularDependencyException(Exception):

    def __init__(self, dependency: Type):
        super().__init__(f'Circular dependency "{dependency.__qualname__}"')


class UnspecifiedDependencyTypeException(Exception):

    def __init__(self, message: str = String.EMPTY):
        super().__init__(message)


class MultipleProviderForDependencyResolution(Exception):

    def __init__(self, dependency: Type):
        super().__init__(f'More than one provider method for dependency "{dependency.__qualname__}" detected')


class DuplicateSingletonInstanceException(Exception):

    def __init__(self, singleton_dependency: Type):
        super().__init__(f'Duplicate singleton instance "{singleton_dependency.__qualname__}" detected')


class TargetTypeNotSpecifiedException(Exception):
    pass


class TargetElementTypeMisMatchException(Exception):

    def __init__(self,
                 target_decorator: Callable,
                 declared_element_type: ElementType,
                 received_element_type: ElementType):

        super().__init__(
            f'Target decorator: "{target_decorator.__qualname__}" placed wrongfully.{String.NEXT_LINE}'
            f'Should be applied to of type: "{declared_element_type.upper()}" only.{String.NEXT_LINE}'
            f'But found to be applied to of type: "{received_element_type.upper()}".{String.NEXT_LINE}'
        )



from strenum import StrEnum


class ElementType(StrEnum):
    CLASS = 'class'
    FUNCTION = 'function'
    STATIC_METHOD = FUNCTION  # haven't found any way to programmatically distinguish these two
    INSTANCE_METHOD = 'instance_method'
    CLASS_METHOD = 'class_method'
    UNDETERMINED = 'undetermined'



from functools import wraps
from typing import Callable, Any

from .target import target
from ..constants.element_type import ElementType
from ..resource_scanner.resource_scanner import RepositoryScanner


@target(ElementType.FUNCTION)
def pinjet_application(function: Callable) -> Callable:

    @wraps(function)
    def decorated_function(*args, **kwargs) -> Any:

        RepositoryScanner.scan_repository(function)

        result: Any = function(*args, **kwargs)

        return result

    return decorated_function



import inspect
from inspect import Signature, Parameter
from typing import Type, Callable, Dict, List

from .target import target
from ..bindings.provider_registry import ProviderMappings
from ..constants.element_type import ElementType
from ..exception.exceptions import (
    UnspecifiedDependencyTypeException,
    MultipleProviderForDependencyResolution,
)
from ..annotations.injectable import injectable
from ..types.resolution_procedure import ResolutionProcedure


@target(ElementType.INSTANCE_METHOD)
def provides(function: Callable) -> Callable:

    properties: Dict = vars(function)

    decorator_list: List[Callable] = properties.setdefault('applied_decorators', [])

    if provides not in decorator_list:
        decorator_list.append(provides)

    return function


@target(ElementType.CLASS)
def provider(clazz: Type) -> Type:

    injectable(clazz)

    members = inspect.getmembers(clazz)

    for name, member in members:

        if inspect.isfunction(member) is not True:
            continue

        member_properties: Dict = vars(member)
        applied_decorator_list: List[Callable] = member_properties.setdefault('applied_decorators', [])

        if provides in applied_decorator_list:

            signature: Signature = inspect.signature(member)
            target_instance_type: Type = signature.return_annotation

            if target_instance_type is None or target_instance_type is Parameter.empty:
                raise UnspecifiedDependencyTypeException(f'Missing return type in Provider method: {name}')

            if ProviderMappings.contains_resolver_for(target_instance_type):
                raise MultipleProviderForDependencyResolution(target_instance_type)

            resolution_procedure: ResolutionProcedure = ResolutionProcedure(source_class=clazz,
                                                                            source_function_name=name)

            ProviderMappings.put(target_instance_type, resolution_procedure)

    return clazz



from typing import Type, Dict

from ..constants.dependency_scope import DependencyScope
from ..types.generic_types import T
from pinjet.common.constants import String


class DependencyMappings:

    def __init__(self):
        if not hasattr(self, '_mappings'):
            self._mappings: Dict[Type, Type] = {}
        if not hasattr(self, '_scope_mappings'):
            self._scope_mappings: Dict[Type, DependencyScope] = {}

    def __new__(cls):

        if not hasattr(cls, 'singleton_instance'):
            setattr(cls, 'singleton_instance', super(DependencyMappings, cls).__new__(cls))
        return getattr(cls, 'singleton_instance')

    @staticmethod
    def __get_singleton_instance() -> 'DependencyMappings':
        return DependencyMappings()

    @staticmethod
    def bind(key: Type, value: Type, scope=DependencyScope.SINGLETON) -> None:
        DependencyMappings.__get_singleton_instance()._mappings[key] = value
        DependencyMappings.__get_singleton_instance()._scope_mappings[key] = scope

    @staticmethod
    def get_binding(key: Type[T]) -> Type[T]:
        return DependencyMappings.__get_singleton_instance()._mappings.get(key)

    @staticmethod
    def get_scope(key: Type) -> DependencyScope:
        return DependencyMappings.__get_singleton_instance()._scope_mappings.get(key)

    @staticmethod
    def contains_binding_for(key: Type) -> bool:
        return True if key in DependencyMappings.__get_singleton_instance()._mappings else False

    def __str__(self) -> str:
        return f'class_mappings {str(self._mappings)}' \
               f'{String.NEXT_LINE}' \
               f'scope_mappings {str(self._scope_mappings)}'



import codecs

from setuptools import setup, find_packages


VERSION = '0.0.1-beta.1'
DESCRIPTION = 'Small like a pin, fast like a jet! A lightweight Dependency Injection library for python'
AUTHOR = 'abetrack3 (Abrar Shahriar Abeed)'
AUTHOR_EMAIL = '<abrarshahriar2361@gmail.com>'

with codecs.open('README.md', encoding='utf-8') as readme_file:
    long_description = '\n' + readme_file.read()
    print(long_description)

setup(
    name='pinjet-core',
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    install_requires=[
        'pinjet-common~=1.0.2'
    ],
    keywords=[
        'python',
        'dependency injection',
        'di framework',
        'ioc',
        'inversion of control',
    ],
)
